# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import importlib.util as mprt
import sys as sstm
from os import sep as PATH_SEPARATOR
from pathlib import Path as path_t
from types import ModuleType
from typing import Any, Callable, Protocol, TypeVar

button_wgt_h = TypeVar("button_wgt_h")
choices_dots_wgt_h = TypeVar("choices_dots_wgt_h")
choices_list_wgt_h = TypeVar("choices_list_wgt_h")
grid_lyt_h = TypeVar("grid_lyt_h")
group_wgt_h = TypeVar("group_wgt_h")
hbox_lyt_h = TypeVar("hbox_lyt_h")
label_wgt_h = TypeVar("label_wgt_h")
library_wgt_h = TypeVar("library_wgt_h")
stack_wgt_h = TypeVar("stack_wgt_h")
stacked_wgt_h = TypeVar("stacked_wgt_h")
text_wgt_h = TypeVar("text_wgt_h")
vbox_lyt_h = TypeVar("vbox_lyt_h")


class _protocol_t(Protocol):
    ALIGNED_HCENTER: Any
    ALIGNED_LEFT: Any
    ALIGNED_RIGHT: Any
    ALIGNED_TOP: Any
    BASE_PALETTE: Any
    COLOR_CYAN: Any
    DIALOG_ACCEPTATION: Any
    DIALOG_ACCEPT_OPEN: Any
    DIALOG_ACCEPT_SAVE: Any
    DIALOG_AUTO_OVERWRITE: Any
    DIALOG_MODE_ANY: Any
    DIALOG_MODE_EXISTING_FILE: Any
    DIALOG_MODE_FOLDER: Any
    FORMAT_RICH: Any
    SELECTABLE_TEXT: Any
    SIZE_EXPANDING: Any
    SIZE_FIXED: Any
    SIZE_MINIMUM: Any
    TAB_POSITION_EAST: Any

    library_wgt_t: Any

    button_wgt_t: Any
    choice_menu_wgt_t: Any
    dot_button_wgt_t: Any
    group_wgt_t: Any
    label_wgt_t: Any
    path_chooser_wgt_t: Any
    scroll_container_t: Any
    stack_wgt_t: Any
    tabs_wgt_t: Any
    text_wgt_t: Any

    hbox_lyt_t: Any
    vbox_lyt_t: Any
    grid_lyt_t: Any

    event_loop_t: Any

    ShowErrorMessage: Callable[..., None]
    ShowMessage: Callable[..., None]


@dtcl.dataclass(repr=False, eq=False)
class backend_t(_protocol_t):
    name: str

    def __post_init__(self) -> None:
        """"""
        base_path = path_t(__file__).parent.parent.parent
        package_path = base_path.parent
        path = base_path / "catalog" / "interface" / "window" / "backend" / self.name
        if not path.is_dir():
            raise ValueError(f"{path}: Invalid backend folder.")

        standard_modules = set(sstm.stdlib_module_names).union(
            sstm.builtin_module_names
        )
        for node in path.rglob("*.py"):
            if node.is_file():
                relative = node.relative_to(package_path)
                name = str(relative.parent / relative.stem).replace(PATH_SEPARATOR, ".")
                spec = mprt.spec_from_file_location(name, node)
                module = mprt.module_from_spec(spec)
                sstm.modules[name] = module
                spec.loader.exec_module(module)

                for name in dir(module):
                    if name[0] == "_":
                        continue

                    element = getattr(module, name)
                    if not (
                        isinstance(element, ModuleType)
                        or (element.__module__[0] == "_")
                        or (element.__module__ in standard_modules)
                    ):
                        setattr(self, name, element)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any

from conf_ini_g.catalog.specification.parameter.choices import choices_t
from conf_ini_g.interface.window.backend import backend_t, choices_list_wgt_h
from conf_ini_g.phase.specification.parameter.main import parameter_t
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class choices_wgt_t:
    library_wgt: choices_list_wgt_h

    @classmethod
    def NewWithDetails(
        cls,
        value: str | None,
        value_type: type_t | choices_t | None,
        backend: backend_t,
        _: parameter_t | None,
        /,
    ) -> choices_wgt_t:
        """
        If value_type does not contain the necessary details, the initial value (if valid) is the only choice, or a unique
        default choice ending with an exclamation point is added.
        """
        output = cls(library_wgt=backend.choice_menu_wgt_t())

        value_is_valid = isinstance(value, str)
        if value_is_valid:
            value = value.strip()

        if value_type is None:
            choices = _DefaultChoices(value, value_is_valid)
        else:
            if isinstance(value_type, type_t):
                annotation = value_type.FirstAnnotationWithAttribute("options")
            else:
                annotation = value_type
            if annotation is None:
                choices = _DefaultChoices(value, value_is_valid)
            else:
                choices = annotation.options

        for choice in choices:
            output.library_wgt.addItem(choice)
        if value_is_valid:
            output.library_wgt.setCurrentText(value)
        else:
            output.library_wgt.setCurrentIndex(0)

        return output

    def __getattr__(self, attribute: str, /) -> Any:
        """
        E.g., used for "SetFunction".
        """
        try:
            output = super(choices_wgt_t, self).__getattr__(attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output


def _DefaultChoices(value: str, value_is_valid: bool, /) -> tuple[str]:
    """"""
    if value_is_valid:
        return (value,)
    else:
        return ("Default!",)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from types import NoneType, UnionType
from typing import Sequence

from conf_ini_g.catalog.interface.window.parameter.directory import WidgetTypeForType
from conf_ini_g.catalog.interface.window.parameter.multitype import multitype_wgt_t
from conf_ini_g.extension.parser.type_hint import hint_tree_t
from conf_ini_g.interface.resource import INI_UNIT_SEPARATOR
from conf_ini_g.interface.window.backend import (
    backend_t,
    choices_list_wgt_h,
    label_wgt_h,
    library_wgt_h,
    text_wgt_h,
)
from conf_ini_g.interface.window.generic import FormattedName
from conf_ini_g.phase.instance.parameter.main import instance_t
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class parameter_t:
    """
    In order to leave the section widget put the name, type, and input widgets of each parameter in columns,
    parameter_t is not a container widget. Instead, it just stores its component widgets for later addition to a layout.
    """

    name: label_wgt_h = dtcl.field(init=False, default=None)
    type: label_wgt_h | choices_list_wgt_h = dtcl.field(init=False, default=None)
    value: library_wgt_h = dtcl.field(init=False, default=None)
    unit: text_wgt_h = dtcl.field(init=False, default=None)
    comment: str = dtcl.field(init=False, default=None)

    @classmethod
    def NewForParameter(
        cls,
        parameter_spec: parameter_spec_t,
        instance: instance_t,
        backend: backend_t,
        /,
    ) -> parameter_t:
        """"""
        output = cls()

        formatted_name = FormattedName(parameter_spec.name, " ")
        comment = f"{formatted_name}\n{parameter_spec.definition}.\n\n{parameter_spec.description}."
        if instance.comment is not None:
            comment += f"\n\n{instance.comment}."
        output.name = backend.label_wgt_t(formatted_name)
        output.name.setToolTip(comment)
        output.comment = comment

        type_ = parameter_spec.type
        value, issues = instance.TypedValue(type_)
        if issues.__len__() > 0:
            # TODO: Show issues in dialog window.
            value = None
        if isinstance(type_, hint_tree_t) and (type_.type is UnionType):
            hint_options = type_.elements
            type_wgt = _TypeSelector(hint_options, backend)
            value_wgt = multitype_wgt_t.NewForHints(
                value,
                hint_options,
                type_wgt,
                backend,
            )
        else:
            type_wgt = backend.label_wgt_t(type_.template_as_str)
            widget_type = WidgetTypeForType(type_)
            value_wgt = widget_type.NewWithDetails(
                value,
                type_,
                backend,
                parameter_spec,
            )
        output.type = type_wgt
        output.value = value_wgt

        if (unit := getattr(instance, "unit", None)) is not None:
            output.unit = backend.text_wgt_t(unit)

        name_style = "padding-right: 5px;"
        if parameter_spec.optional:
            name_style += "color: gray;"
        output.name.setStyleSheet(name_style)
        output.type.setStyleSheet(name_style)

        return output

    def SetVisible(self, visible: bool, /) -> None:
        """"""
        self.name.setVisible(visible)
        self.type.setVisible(visible)
        if self.unit is not None:
            self.unit.setVisible(visible)

    def Text(self) -> str:
        """"""
        text = self.value.Text()

        if self.unit is None:
            return text

        return f"{text}{INI_UNIT_SEPARATOR}{self.unit.Text()}"


def _TypeSelector(
    hints: Sequence[hint_tree_t], backend: backend_t, /
) -> choices_list_wgt_h:
    """"""
    output = backend.choice_menu_wgt_t()

    for hint in hints:
        # Here, it is important that hint trees cannot have an OR-node with an OR-node
        # child so that hint.type has a valid value.
        if hint is NoneType:
            output.addItem("None")
        else:
            output.addItem(hint.type.__name__)

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import re as regx
from types import EllipsisType, UnionType
from typing import Any, Union, get_args, get_origin

from conf_ini_g.extension.type import (
    HintComponents,
    annotation_t,
    any_hint_h,
    complex_hint_additions_h,
    non_complex_hint_h,
    none_t,
    simple_hint_h,
)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class _hint_node_t:
    """
    Leave elements to the tree.
    """
    type: non_complex_hint_h | UnionType | EllipsisType | type[none_t]
    annotations: tuple[annotation_t, ...] = ()

    # TODO: Maybe add these to hint_tree_t also.
    # def __str__(self) -> str:
    #     """"""
    #     return text_t.from_markup(self.__rich__()).plain
    #
    # def __rich__(self) -> str:
    #     """"""
    #     output = [TypeAsRichStr(self)]
    #
    #     names = (_fld.name for _fld in dtcl.fields(self))
    #     for name in names:
    #         value = getattr(self, name)
    #         output.append(f"    {NameValueTypeAsRichStr(name, value, separator='@=@')}")
    #
    #     output = AlignedOnSeparator(output, "@=@", " = ")
    #
    #     return "\n".join(output)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class hint_tree_t(_hint_node_t):
    elements: tuple[hint_tree_t, ...] = None

    @classmethod
    def NewFromTypeHint(
        cls, hint: any_hint_h | complex_hint_additions_h, /
    ) -> hint_tree_t:
        """
        Note that type hints cannot translate into hint trees with an OR-node having a child
        OR-node. For example: str | (int | float) is interpreted as str | int | float. This
        is important when creating a type selector for multi-type parameters since only
        direct child nodes are taken into account for widget creation, so these nodes must
        be types, not an OR subtree.
        """
        if isinstance(hint, annotation_t):
            # This is a common mistake in specifications.
            raise ValueError(
                f'{hint}: Invalid hint type "{type(hint).__name__}"; '
                f"Expected={any_hint_h}."
            )

        # Dealing with complex_hint_additions_h first
        if hint is Ellipsis:
            return cls(type=EllipsisType)
        if hint is None:
            return cls(type=none_t)

        # nnts: Do not use "annotations" since it shadows __future__.annotations.
        raw_hint, nnts = HintComponents(hint)

        if (origin := get_origin(raw_hint)) is None:
            return cls(type=raw_hint, annotations=nnts)

        # Handled types: list, set, tuple, with sets using the dict delimiters { and }.
        if origin is dict:
            raise TypeError(f"{origin.__name__}: Unhandled type.")

        if origin is Union:
            origin = UnionType
        elements = tuple(cls.NewFromTypeHint(_elm) for _elm in get_args(raw_hint))
        return cls(type=origin, annotations=nnts, elements=elements)

    @property
    def all_annotations(self) -> list[annotation_t]:
        """"""
        output = list(self.annotations)

        if self.elements is not None:
            for element in self.elements:
                output.extend(element.all_annotations)

        return output

    @property
    def template(self) -> simple_hint_h | dict[int, Any] | None:
        """"""
        if self.type is none_t:
            return None

        if self.type is UnionType:
            return {_key: _elm.template for _key, _elm in enumerate(self.elements)}

        if self.elements is None:
            return self.type

        return self.type(_elm.template for _elm in self.elements)

    @property
    def template_as_str(self) -> str:
        """"""
        output = (
            str(self.template)
            .replace("<class '", "")
            .replace("'>", "")
            .replace(str(Ellipsis), "...")
        )
        output = regx.sub("{\d: ", "{", output, flags=regx.ASCII)
        output = regx.sub(", \d:", " |", output, flags=regx.ASCII)

        return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as path_t
from types import EllipsisType, UnionType
from typing import Any, Iterable

from conf_ini_g.extension.parser.type_hint import hint_tree_t
from conf_ini_g.extension.type import any_hint_h, none_t, simple_hint_h


class _not_a_leaf_node_t:
    pass


_NOT_A_LEAF_NODE = _not_a_leaf_node_t()


@dtcl.dataclass(slots=True, repr=False, eq=False)
class _value_node_t:
    """
    Leave elements to the tree.
    leaf_value: None for non-leaf nodes; Actual value (before or after cast) for leaves.
    """

    type: simple_hint_h = None
    leaf_value: Any | None = _NOT_A_LEAF_NODE


@dtcl.dataclass(slots=True, repr=False, eq=False)
class value_tree_t(_value_node_t):
    elements: tuple[value_tree_t, ...] = None

    @classmethod
    def NewFromValue(cls, value: Any, /) -> value_tree_t:
        """"""
        if isinstance(value, Iterable) and not isinstance(value, str):
            elements = tuple(cls.NewFromValue(_elm) for _elm in value)
            return cls(type=type(value), elements=elements)

        return cls(leaf_value=value)

    def _RebuiltValue(self) -> Any:
        """"""
        if self.leaf_value is _NOT_A_LEAF_NODE:
            return self.type(_elm._RebuiltValue() for _elm in self.elements)

        return self.leaf_value

    def CastValue(
        self, hint_tree: hint_tree_t, /, *, only_check_validity: bool = False
    ) -> tuple[Any, list[str]] | list[str]:
        """"""
        issues = self._CastToHint(hint_tree)
        if issues.__len__() > 0:
            if only_check_validity:
                return issues
            else:
                return None, issues

        if only_check_validity:
            return []
        else:
            return self._RebuiltValue(), []

    def _CastToHint(self, hint_node: hint_tree_t, /) -> list[str]:
        """
        Returned value=the value tree has been successfully cast to the hint tree
        specification, or not.
        """
        hn_type = hint_node.type
        hn_elements = hint_node.elements

        if hn_type is Any:
            return self._CompliesWithAnnotations(hint_node)

        if hn_type is none_t:
            # None is not supposed to have annotations. They are ignored if it does.
            if self.leaf_value is None:
                output = []
            else:
                output = [f"{self.leaf_value}: Invalid value; Expected=None."]
            return output

        if hn_type is UnionType:
            output = []
            for element in hn_elements:
                issues = self._CastToHint(element)
                if issues.__len__() > 0:
                    output.extend(issues)
                else:
                    return []
            return output

        if self.elements is None:
            if hn_elements is None:
                if not isinstance(self.leaf_value, hn_type):
                    # Dealing with "equivalent" types, such as "str" and "pathlib.Path".
                    if issubclass(hn_type, path_t):
                        if isinstance(self.leaf_value, str):
                            self.leaf_value = hn_type(self.leaf_value)
                        else:
                            return [
                                f"{self.leaf_value}: Cannot be cast to type "
                                f'"{hn_type.__name__}".'
                            ]
                    else:
                        return [
                            f"{self.leaf_value}: Invalid value type "
                            f'"{type(self.leaf_value).__name__}"; '
                            f'Expected="{hn_type.__name__}".'
                        ]

                return self._CompliesWithAnnotations(hint_node)
            else:
                return [
                    f"{self.leaf_value}: Not a container value; "
                    f"Expected=Container with template {hint_node.template_as_str}."
                ]

        # For then on, self.elements is not None.

        if hn_elements is None:
            # The type hint does not fully specify valid values, so anything is valid.
            if not issubclass(self.type, hn_type):
                rebuilt = self._RebuiltValue()
                try:
                    _ = hn_type(rebuilt)
                except:
                    return [
                        f"{rebuilt}: Cannot be cast to type " f'"{hn_type.__name__}".'
                    ]

            return self._CompliesWithAnnotations(hint_node)

        n_value_elements = self.elements.__len__()
        n_hint_elements = hn_elements.__len__()
        has_ellipsis = (n_hint_elements == 2) and (hn_elements[1].type is EllipsisType)
        should_fake_ellipsis = (n_hint_elements == 1) and issubclass(
            hn_type, (list, set)
        )
        if (
            has_ellipsis
            or should_fake_ellipsis
            or (n_value_elements == n_hint_elements)
        ):
            if has_ellipsis or should_fake_ellipsis:
                adjusted_hn_elements = n_value_elements * (hn_elements[0],)
            else:
                adjusted_hn_elements = hn_elements
            issues = []
            for value_elm, hint_elm in zip(self.elements, adjusted_hn_elements):
                value_elm: value_tree_t
                issues.extend(value_elm._CastToHint(hint_elm))
            if issues.__len__() > 0:
                return issues

            return self._CompliesWithAnnotations(hint_node)

        return [
            f"{self.leaf_value}: Invalid container value; "
            f"Expected=Value following template {hint_node.template_as_str}."
        ]

    def _CompliesWithAnnotations(self, hint_node: hint_tree_t, /) -> list[str]:
        """"""
        output = []

        for annotation in hint_node.annotations:
            output.extend(annotation.ValueIsCompliant(self._RebuiltValue()))

        return output


def CastValue(
    value: Any, hint: any_hint_h | hint_tree_t, /, *, only_check_validity: bool = False
) -> tuple[Any, list[str]] | list[str]:
    """"""
    value_tree = value_tree_t.NewFromValue(value)
    if not isinstance(hint, hint_tree_t):
        hint = hint_tree_t.NewFromTypeHint(hint)

    return value_tree.CastValue(hint, only_check_validity=only_check_validity)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from types import EllipsisType, GenericAlias, UnionType
from typing import Annotated, Any, ClassVar, Sequence

from rich.text import Text as text_t

from conf_ini_g.extension.string import AlignedOnSeparator

# --- Unannotated hints
# Atoms
generic_hint_h = Any
simple_hint_h = type[Any]
complex_hint_h = GenericAlias | UnionType
# Grouping
non_complex_hint_h = generic_hint_h | simple_hint_h
raw_hint_h = non_complex_hint_h | complex_hint_h
# --- Annotated hints
annotated_hint_t = type(Annotated[object, None])
# --- [Un]Annotated hints
any_hint_h = raw_hint_h | annotated_hint_t
# --- Complex hints additional components:
#     - EllipsisType for GenericAlias, as in: tuple[int, ...]
#     - None for UnionType, as in: int | None
complex_hint_additions_h = EllipsisType | None


class none_t:  # To play the role of NoneType
    pass


@dtcl.dataclass(slots=True, repr=False, eq=False)
class annotation_t:
    ACCEPTED_TYPES: ClassVar[tuple[non_complex_hint_h, ...]] = (Any,)

    @classmethod
    def NewAnnotatedType(cls, *args, **kwargs) -> annotated_hint_t:
        """
        Recommendation: Should only be implemented if ACCEPTED_TYPES contains a single
        type, thus avoiding to use:
        Annotated[there_is_no_other_choice_then_single_accepted_type, annotation_t(...)]
        and using:
        annotation_t.NewAnnotatedType(...)
        instead.
        """
        raise NotImplementedError

    def Issues(self) -> list[str]:
        """"""
        return []

    def ValueIsCompliant(self, _: Any, /) -> list[str]:
        """"""
        return []

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [TypeAsRichStr(self)]

        names = (_fld.name for _fld in dtcl.fields(self))
        for name in names:
            value = getattr(self, name)
            output.append(f"    {NameValueTypeAsRichStr(name, value, separator='@=@')}")

        output = AlignedOnSeparator(output, "@=@", " = ")

        return "\n".join(output)


def IsAnnotated(hint: any_hint_h, /) -> bool:
    """"""
    return isinstance(hint, annotated_hint_t)


def TypeOfAnnotatedHint(annotated_hint: annotated_hint_t, /) -> raw_hint_h:
    """"""
    return annotated_hint.__args__[0]


def AnnotationsOfAnnotatedHint(annotated_hint: annotated_hint_t, /) -> tuple[Any, ...]:
    """"""
    output = tuple(annotated_hint.__metadata__)
    if all(isinstance(_elm, annotation_t) for _elm in output):
        return output

    raise ValueError(
        f'{output}: Not all elements are of type "{annotation_t.__name__}".'
    )


def HintComponents(hint: any_hint_h, /) -> tuple[raw_hint_h, tuple[Any, ...]]:
    """"""
    if IsAnnotated(hint):
        return TypeOfAnnotatedHint(hint), AnnotationsOfAnnotatedHint(hint)

    return hint, ()


def TypeAsRichStr(value: Any, /, *, prefix: str = "") -> str:
    """
    Richer alternative (however, causes troubles with packages like TensorFlow):
    Additional parameter: relative_to_home: bool = True
    from conf_ini_g.extension.python import SpecificationPath
    return (
        f"[bold magenta]{type(instance).__name__}[/]"
        f"[gray]@"
        f"{SpecificationPath(type(instance), relative_to_home=relative_to_home)}:[/]"
    )
    """
    return f"[yellow]{prefix}{type(value).__name__}[/]"


def NameValueTypeAsRichStr(name: str, value: Any, /, *, separator: str = "=") -> str:
    """"""
    formatted_type = TypeAsRichStr(value, prefix=":")
    if isinstance(value, Sequence) and (value.__len__() == 0):
        value = "[cyan]<empty>[/]"

    return f"[blue]{name}{separator}[/]{value}{formatted_type}"



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as pl_path_t
from typing import Callable, Sequence, TypeVar

from conf_ini_g.catalog.specification.parameter.path import path_t
from conf_ini_g.extension.path import any_path_h
from conf_ini_g.interface.window.backend import backend_t

"""
valid_types: {"Type": "extension", "Type": ("extension", "extension",...), ...}
filter: "Image files (*.png *.xpm *.jpg);Text files (*.txt);Any files (*)"
"""


path_chooser_wgt_h = TypeVar("path_chooser_wgt_h")
valid_types_h = dict[str, str | Sequence[str]]
document_selection_fct_h = Callable[..., pl_path_t | None]


def NewSelectedInputDocument(
    title: str,
    caption: str,
    backend: backend_t,
    /,
    *,
    mode: path_t.TARGET_TYPE = path_t.TARGET_TYPE.any,
    valid_types: valid_types_h = None,
    start_folder: any_path_h = None,
    initial_selection: any_path_h = None,
) -> pl_path_t | None:
    """"""
    extension_filter, _ = _AllowedTypesElements(valid_types)
    check_existence = False
    if mode is path_t.TARGET_TYPE.document:
        dialog_mode = backend.DIALOG_MODE_EXISTING_FILE
    elif mode is path_t.TARGET_TYPE.folder:
        dialog_mode = backend.DIALOG_MODE_FOLDER
    else:
        # TODO: Check if that allows to select a folder (documentation says "The name of
        #     a file, whether it exists or not."). So a priori, no. But then how can we
        #     allow selection of either a file or a folder?
        dialog_mode = backend.DIALOG_MODE_ANY
        check_existence = True

    while True:
        dialog = _GenericDocumentDialog(
            title, caption, extension_filter, start_folder, initial_selection, backend
        )
        dialog.setAcceptMode(backend.DIALOG_ACCEPT_OPEN)
        dialog.setFileMode(dialog_mode)

        output = _SelectedDocument(dialog, backend)
        if output is None:
            return None
        if check_existence and not output.exists():
            backend.ShowErrorMessage(f"{output}: Nonexistent file or folder")
            start_folder = _StartFolderFromFolder(output)
            initial_selection = None
        else:
            # The file dialog does not allow to select either a file or a folder. So the
            # solution here is to select a file, and if a folder was needed, take the
            # parent.
            if (mode is path_t.TARGET_TYPE.folder) and output.is_file():
                output = output.parent
            return output


def NewSelectedOutputDocument(
    title: str,
    caption: str,
    backend: backend_t,
    /,
    *,
    mode: path_t.TARGET_TYPE = path_t.TARGET_TYPE.any,
    valid_types: valid_types_h = None,
    auto_overwrite: bool = False,
    start_folder: any_path_h = None,
    initial_selection: any_path_h = None,
) -> pl_path_t | None:
    """"""
    extension_filter, allowed_extensions = _AllowedTypesElements(valid_types)
    while True:
        dialog = _GenericDocumentDialog(
            title, caption, extension_filter, start_folder, initial_selection, backend
        )
        dialog.setAcceptMode(backend.DIALOG_ACCEPT_SAVE)
        dialog.setFileMode(backend.DIALOG_MODE_ANY)
        if auto_overwrite:
            dialog.setOption(backend.DIALOG_AUTO_OVERWRITE)

        output = _SelectedDocument(dialog, backend)
        if output is None:
            return None
        # The file dialog does not allow to select either a file or a folder. So the solution here is to select a file,
        # and if a folder was needed, take the parent. See (*) below.
        if (mode is path_t.TARGET_TYPE.folder) and output.exists() and output.is_file():
            output = output.parent

        erroneous_selection = False
        if output.exists():
            if (mode is path_t.TARGET_TYPE.document) and not output.is_file():
                backend.ShowErrorMessage(f"{output}: Not of regular file")
                erroneous_selection = True
            # Unnecessary due to (*) above
            # elif (mode == "folder") and not output.is_dir():
            #     ShowErrorMessage(f"{output}: Not a folder")
            #     erroneous_selection = True

        if not erroneous_selection:
            if ("*" in allowed_extensions) or (
                output.suffix.lower()[1:] in allowed_extensions
            ):
                return output
            else:
                backend.ShowErrorMessage(f"{output}: Extension is not valid")

        start_folder = _StartFolderFromFolder(output)
        initial_selection = None


def _AllowedTypesElements(
    valid_types: valid_types_h | None, /
) -> tuple[str, tuple[str]]:
    """"""
    if valid_types is None:
        return "Any file or folder (*)", ("*",)

    types = []
    extensions = []
    filters = []
    for _type, extension in valid_types.items():
        types.append(_type)
        if isinstance(extension, str):
            extensions.append(extension)
            if extension == "*":
                new_extensions = ("*",)
            else:
                new_extensions = (f"*.{extension}",)
        else:
            extensions.extend(extension)
            new_extensions = tuple(f"*.{_ext}" for _ext in extension)
        filters.append(f"{_type.title()} ({' '.join(new_extensions)})")

    return ";".join(filters), tuple(extensions)


def _StartFolderFromFolder(folder: pl_path_t, /) -> pl_path_t | None:
    """"""
    output = folder

    root = folder.root
    while (output != root) and (not output.exists()):
        output = output.parent
    if output == root:
        output = None

    return output


def _GenericDocumentDialog(
    title: str,
    caption: str,
    extension_filter: str,
    start_folder: any_path_h,
    initial_selection: any_path_h,
    backend: backend_t,
    /,
) -> path_chooser_wgt_h:
    """"""
    output = backend.path_chooser_wgt_t(caption, extension_filter=extension_filter)
    output.setWindowTitle(title)
    if start_folder is not None:
        output.setDirectory(str(start_folder))
    if initial_selection is not None:
        output.selectFile(str(initial_selection))

    return output


def _SelectedDocument(
    dialog: path_chooser_wgt_h, backend: backend_t, /
) -> pl_path_t | None:
    """"""
    status = dialog.RunAndGetClosingStatus()
    if status == backend.DIALOG_ACCEPTATION:
        return pl_path_t(dialog.SelectedFile())

    return None



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from pathlib import Path as pl_path_t
from typing import Annotated, Any, Callable, Iterator, Sequence

from conf_ini_g.catalog.interface.window.parameter.boolean import boolean_wgt_t
from conf_ini_g.catalog.interface.window.parameter.path import path_wgt_t
from conf_ini_g.catalog.specification.parameter.boolean import boolean_t
from conf_ini_g.catalog.specification.parameter.path import path_t
from conf_ini_g.catalog.specification.parameter.path import path_t as path_annotation_t
from conf_ini_g.interface.resource import INI_COMMENT_MARKER
from conf_ini_g.interface.storage.config import SaveRawConfigToINIDocument
from conf_ini_g.interface.window.backend import (
    backend_t,
    button_wgt_h,
    grid_lyt_h,
    hbox_lyt_h,
    library_wgt_h,
    vbox_lyt_h,
)
from conf_ini_g.interface.window.component.section import (
    controlled_section_t,
    section_t,
)
from conf_ini_g.interface.window.path_chooser import NewSelectedOutputDocument
from conf_ini_g.phase.instance.config.main import config_t as config_instance_t
from conf_ini_g.phase.raw.config import AsStr, ini_config_h, raw_config_h
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.parameter.type import type_t
from conf_ini_g.phase.specification.section.main import section_t as section_spec_t


@dtcl.dataclass(repr=False, eq=False)
class config_t:
    """
    The class cannot use slots because it disables weak referencing, which is required.
    See error message below when using slots:
    TypeError: cannot create weak reference to 'config_t' object
    [...]
    File "[...]conf_ini_g/catalog/interface/window/backend/pyqt5/widget/choices.py", line 41, in SetFunction
        self.clicked.connect(function)
        │                    └ <bound method config_t.ToogleAdvancedMode of <conf_ini_g.interface.window.config.config_t object at [...]>>
        └ <conf_ini_g.catalog.interface.window.backend.pyqt5.widget.choices.dot_button_wgt_t object at [...]>

    Widget might not cooperate well with list, in which case Python raises the
    following exception: TypeError: multiple bases have instance lay-out conflict
    To be safe, "sections" is a field instead of being part of the class definition.
    """

    sections: dict[str, section_t | controlled_section_t] = dtcl.field(
        init=False, default=None
    )
    instance: config_instance_t
    backend: backend_t
    library_wgt: library_wgt_h
    _action_button: button_wgt_h = dtcl.field(init=False, default=None)
    Action: Callable[[raw_config_h], None] = None
    _reference_keeper: tuple[library_wgt_h, ...] = None

    @classmethod
    def NewFromConfig(
        cls,
        title: str | None,
        instance: config_instance_t,
        backend: backend_t,
        /,
        *,
        advanced_mode: bool = False,
        action: tuple[str, Callable[[raw_config_h], None]] = None,
    ) -> config_t:
        """"""
        if action is None:
            kwargs = {}
        else:
            kwargs = {"Action": action[1]}
        output = cls(
            instance=instance,
            backend=backend,
            library_wgt=backend.library_wgt_t(),
            **kwargs,
        )

        # --- Top-level widgets
        title_lyt, widget_1 = _TitleLayout(title, instance, output)
        advanced_mode_lyt, widget_2 = _AdvancedModeLayout(advanced_mode, output)
        button_lyt, widgets = _ActionButtonsLayout(
            output, action, instance.ini_path is not None
        )
        output._reference_keeper = (widget_1, widget_2, *widgets)

        # --- Sections
        categories = {}
        sections = {}
        controlled_sections = []

        for section_spec in instance.specification:
            if (controller := section_spec.controller) is None:
                section = section_t.NewForSection(
                    section_spec, instance[section_spec.name], backend
                )
            else:
                section = controlled_section_t.NewForSection(
                    section_spec,
                    controller,
                    instance.GetValueOfController(controller),
                    instance[section_spec.name],
                    backend,
                )
                if section is not None:
                    controlled_sections.append((section, section_spec))
            if section is None:
                continue

            sections[section_spec.name] = section

            if (category := section_spec.category) not in categories:
                contents = backend.library_wgt_t()
                scroll_area = backend.scroll_container_t.NewForWidget(contents)
                layout = backend.vbox_lyt_t()
                contents.setLayout(layout)
                categories[category] = (layout, scroll_area)

            layout = categories[category][0]
            layout.addWidget(section.library_wgt)

        output.sections = sections

        if categories.__len__() > 1:
            category_selector = backend.tabs_wgt_t()
            for category, (_, scroll_area) in categories.items():
                category_selector.addTab(scroll_area, category)
        else:
            category = tuple(categories.keys())[0]
            category_selector = categories[category][1]

        for section, section_spec in controlled_sections:
            controller = section_spec.controller
            parameter = output[controller.section].parameters[controller.parameter]
            value_wgt = parameter.value
            if hasattr(value_wgt, "SetFunction"):
                value_wgt.SetFunction(section.page_stack.setCurrentIndex)
            else:
                backend.ShowErrorMessage(
                    f"{controller.section}.{controller.parameter}: "
                    f'Controller has no "SetFunction" method; Disabling control.'
                )

        # --- Layout...
        layout = backend.grid_lyt_t()
        if title_lyt is None:
            first_available_row = 0
        else:
            layout.addLayout(title_lyt, 0, 0, 1, 1)
            first_available_row = 1
        layout.addWidget(category_selector, first_available_row, 0, 1, 1)
        layout.addLayout(advanced_mode_lyt, first_available_row + 1, 0, 1, 1)
        layout.addLayout(button_lyt, first_available_row + 2, 0, 1, 1)

        output.library_wgt.setLayout(layout)
        # --- ...Layout

        output.ToogleAdvancedMode(advanced_mode)

        return output

    def SectionActiveParameterSpec(
        self, section_spec: section_spec_t, /
    ) -> Sequence[parameter_spec_t]:
        """"""
        if (controller := section_spec.controller) is None:
            output = section_spec
        else:
            controller = self[controller.section].parameters[controller.parameter]
            output = section_spec.ActiveParameters(controller.Text())

        return output

    def ToogleAdvancedMode(self, advanced_mode: bool, /) -> None:
        """"""
        for section_name, section in self.sections.items():
            section_spec = self.instance.specification[section_name]
            if section_spec.basic:
                should_check_parameters = True
            elif advanced_mode:
                section.library_wgt.setVisible(True)
                should_check_parameters = True
            else:
                section.library_wgt.setVisible(False)
                should_check_parameters = False

            if should_check_parameters:
                parameters = section.active_parameters.values()
                parameter_specs = self.SectionActiveParameterSpec(section_spec)
                for parameter, parameter_spec in zip(parameters, parameter_specs):
                    if not parameter_spec.basic:
                        if advanced_mode:
                            parameter.setVisible(True)
                        else:
                            parameter.setVisible(False)

    def SynchronizeInstance(self) -> list[str]:
        """"""
        for section_name, section in self.sections.items():
            section_spec = self.instance.specification[section_name]
            parameters = section.active_parameters.values()
            parameter_specs = self.SectionActiveParameterSpec(section_spec)
            for parameter, parameter_spec in zip(parameters, parameter_specs):
                if parameter.unit is None:
                    unit_kwarg = {}
                else:
                    unit_kwarg = {"unit": parameter.unit.Text()}
                instance = self.instance[section_spec.name][parameter_spec.name]
                instance.SetStrValue(
                    parameter.Text(),
                    INI_COMMENT_MARKER,
                    **unit_kwarg,
                )

        return self.instance.Issues()

    def AsINIConfig(self) -> ini_config_h | None:
        """"""
        issues = self.SynchronizeInstance()
        if issues.__len__() > 0:
            output = None
            self.backend.ShowErrorMessage("\n".join(issues), parent=self)
        else:
            output = self.instance.AsINIConfig()

        return output

    def ShowInINIFormat(self) -> None:
        """"""
        config = self.AsINIConfig()
        if config is None:
            return

        config = AsStr(config, color="html")
        self.backend.ShowMessage("INI Config", "<tt>" + config + "<tt/>")

    def SaveConfig(self, new_ini: bool, /) -> None:
        """"""
        if new_ini:
            path = NewSelectedOutputDocument(
                "Save Config As",
                "Save Config As",
                self.backend,
                mode=path_t.TARGET_TYPE.document,
                valid_types={"Config files": ("ini", "INI")},
            )
        else:
            path = self.instance.ini_path  # Will overwrite current INI document

        if path is not None:
            config = self.AsINIConfig()
            error = SaveRawConfigToINIDocument(config, path)
            if error is None:
                self.instance.ini_path = path
            else:
                self.backend.ShowErrorMessage(error, parent=self)

    def LaunchAction(self) -> None:
        """"""
        issues = self.SynchronizeInstance()
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
        else:
            raw_config, issues = self.instance.AsRawConfig()
            if issues.__len__() > 0:
                self.backend.ShowErrorMessage(
                    "\n".join(issues), parent=self.library_wgt
                )
            else:
                self._action_button.setEnabled(False)
                try:
                    self.Action(raw_config)
                except Exception as exception:
                    self.backend.ShowErrorMessage(
                        str(exception), parent=self.library_wgt
                    )
                self._action_button.setEnabled(True)

    def __getattr__(self, attribute: str, /) -> Any:
        """
        E.g., used for "show".
        """
        try:
            output = super().__getattr__(attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output

    def __contains__(self, key: str, /) -> bool:
        """"""
        return key in self.sections

    def __getitem__(self, key: str, /) -> section_t | controlled_section_t:
        """"""
        return self.sections[key]

    def __iter__(self) -> Iterator[str]:
        """"""
        return self.sections.keys()


def _TitleLayout(
    title: str | None, instance: config_instance_t, config: config_t, /
) -> tuple[vbox_lyt_h, path_wgt_t]:
    """"""
    layout = config.backend.vbox_lyt_t()

    if title is None:
        title = "Conf-INI-g"
    title = (
        f'<h1 style="color: blue">{title}</h1>'
        f"<b><font face=monospace>SPEC:</font></b> {instance.specification.spec_path}"
    )
    title_wgt = config.backend.label_wgt_t(title)
    title_wgt.setAlignment(config.backend.ALIGNED_LEFT)

    if instance.ini_path is None:
        ini_path_wgt = path_layout = None
    else:
        ini_path_wgt = path_wgt_t.NewWithDetails(
            pl_path_t(instance.ini_path),
            path_annotation_t(
                target_type=path_annotation_t.TARGET_TYPE.document, is_input=True
            ),
            config.backend,
            None,
            editable=False,
        )
        label_wgt = config.backend.label_wgt_t(
            "<b><font face=monospace>INI: </font></b>"
        )
        label_wgt.setSizePolicy(config.backend.SIZE_FIXED, config.backend.SIZE_FIXED)
        path_layout = config.backend.hbox_lyt_t()
        path_layout.addWidget(label_wgt)
        path_layout.addWidget(ini_path_wgt.library_wgt)

    layout.addWidget(title_wgt)
    if path_layout is not None:
        layout.addLayout(path_layout)

    return layout, ini_path_wgt


def _AdvancedModeLayout(
    advanced_mode: bool, config: config_t, /
) -> tuple[hbox_lyt_h, boolean_wgt_t]:
    """"""
    layout = config.backend.hbox_lyt_t()

    annotated_type = Annotated[bool, boolean_t(mode=boolean_t.MODE.on_off)]
    value_type = type_t.NewFromTypeHint(annotated_type)
    boolean = boolean_wgt_t.NewWithDetails(
        advanced_mode,
        value_type,
        config.backend,
        None,
    )
    boolean.true_btn.SetFunction(config.ToogleAdvancedMode)

    layout.addWidget(config.backend.label_wgt_t("<i>Advanced Mode</i>"))
    layout.addWidget(boolean.library_wgt)

    return layout, boolean


def _ActionButtonsLayout(
    config: config_t,
    action: tuple[str, Callable[[raw_config_h], None]] | None,
    has_ini_document: bool,
    /,
) -> tuple[grid_lyt_h, Sequence[button_wgt_h, ...]]:
    """"""
    layout = config.backend.grid_lyt_t()

    buttons = []
    geometries = []

    button = config.backend.button_wgt_t("Show in INI format")
    button.SetFunction(config.ShowInINIFormat)
    buttons.append(button)
    geometries.append((0, 0, 1, 2))

    button = config.backend.button_wgt_t("Save Config As")
    button.SetFunction(lambda: config.SaveConfig(True))
    buttons.append(button)
    if has_ini_document:
        geometries.append((1, 0, 1, 1))

        button = config.backend.button_wgt_t("Save/Overwrite Config")
        button.SetFunction(lambda: config.SaveConfig(False))
        buttons.append(button)
        geometries.append((1, 1, 1, 1))
    else:
        geometries.append((1, 0, 1, 2))

    if action is None:
        label = "Close"
        Function = config.library_wgt.close
    else:
        label = action[0]
        Function = config.LaunchAction

    button = config.backend.button_wgt_t(label)
    button.SetFunction(Function)
    buttons.append(button)
    geometries.append((2, 0, 1, 2))

    config._action_button = button

    for button, geometry in zip(buttons, geometries):
        layout.addWidget(button, *geometry)
    layout.setContentsMargins(0, 0, 0, 0)

    return layout, buttons
